@@ -1,27 +1,30 @@
 import base64
+import logging
 import re
 from functools import partial
 
 from Crypto.Cipher import Blowfish
 
-from streamlink import PluginError
 from streamlink.compat import bytes, is_py3
 from streamlink.plugin import Plugin, PluginArguments, PluginArgument
-from streamlink.plugin.api import http
 from streamlink.plugin.api import useragents
 from streamlink.plugin.api import validate
+from streamlink.plugin.api.utils import itertags
 from streamlink.stream import HLSStream
 from streamlink.stream import HTTPStream
 from streamlink.stream.ffmpegmux import MuxedStream
 from streamlink.utils import parse_xml
 
+log = logging.getLogger(__name__)
+
 
 class ZTNRClient(object):
     base_url = "http://ztnr.rtve.es/ztnr/res/"
     block_size = 16
 
-    def __init__(self, key):
+    def __init__(self, key, session):
         self.cipher = Blowfish.new(key, Blowfish.MODE_ECB)
+        self.session = session
 
     @classmethod
     def pad(cls, data):
@@ -42,7 +45,7 @@ class ZTNRClient(object):
         return self.unpad(self.cipher.decrypt(base64.b64decode(data, altchars=b"-_")))
 
     def request(self, data, *args, **kwargs):
-        res = http.get(self.base_url + self.encrypt(data), *args, **kwargs)
+        res = self.session.http.get(self.base_url + self.encrypt(data), *args, **kwargs)
         return self.decrypt(res.content)
 
     def get_cdn_list(self, vid, manager="apedemak", vtype="video", lang="es", schema=None):
@@ -55,9 +58,8 @@ class ZTNRClient(object):
 
 class Rtve(Plugin):
     secret_key = base64.b64decode("eWVMJmRhRDM=")
-    content_id_re = re.compile(r'data-id\s*=\s*"(\d+)"')
     url_re = re.compile(r"""
-        https?://(?:www\.)?rtve\.es/(?:directo|noticias|television|deportes|alacarta|drmn)/.*?/?
+        https?://(?:www\.)?rtve\.es/(?:directo|infantil|noticias|television|deportes|alacarta|drmn)/.*?/?
     """, re.VERBOSE)
     cdn_schema = validate.Schema(
         validate.transform(partial(parse_xml, invalid_char_entities=True)),
@@ -114,21 +116,24 @@ class Rtve(Plugin):
 
     def __init__(self, url):
         Plugin.__init__(self, url)
-        self.zclient = ZTNRClient(self.secret_key)
-        http.headers = {"User-Agent": useragents.SAFARI_8}
+        self.session.http.headers = {"User-Agent": useragents.SAFARI_8}
+        self.zclient = ZTNRClient(self.secret_key, self.session)
 
     def _get_content_id(self):
-        res = http.get(self.url)
-        m = self.content_id_re.search(res.text)
-        return m and int(m.group(1))
+        res = self.session.http.get(self.url)
+        for div in itertags(res.text, "div"):
+            if div.attributes.get("data-id"):
+                return int(div.attributes.get("data-id"))
+        else:
+            log.error("Failed to get content_id")
 
     def _get_subtitles(self, content_id):
-        res = http.get(self.subtitles_api.format(id=content_id))
-        return http.json(res, schema=self.subtitles_schema)
+        res = self.session.http.get(self.subtitles_api.format(id=content_id))
+        return self.session.http.json(res, schema=self.subtitles_schema)
 
     def _get_quality_map(self, content_id):
-        res = http.get(self.video_api.format(id=content_id))
-        data = http.json(res, schema=self.video_schema)
+        res = self.session.http.get(self.video_api.format(id=content_id))
+        data = self.session.http.json(res, schema=self.video_schema)
         qmap = {}
         for item in data["qualities"]:
             qname = {"MED": "Media", "HIGH": "Alta", "ORIGINAL": "Original"}.get(item["preset"], item["preset"])
@@ -139,7 +144,7 @@ class Rtve(Plugin):
         streams = []
         content_id = self._get_content_id()
         if content_id:
-            self.logger.debug("Found content with id: {0}", content_id)
+            log.debug("Found content with id: {0}", content_id)
             stream_data = self.zclient.get_cdn_list(content_id, schema=self.cdn_schema)
             quality_map = None
 
@@ -148,10 +153,10 @@ class Rtve(Plugin):
                     if ".m3u8" in url:
                         try:
                             streams.extend(HLSStream.parse_variant_playlist(self.session, url).items())
-                        except (IOError, OSError):
-                            self.logger.debug("Failed to load m3u8 url: {0}", url)
-                    elif ((url.endswith("mp4") or url.endswith("mov") or url.endswith("avi")) and
-                                  http.head(url, raise_for_status=False).status_code == 200):
+                        except (IOError, OSError) as err:
+                            log.error(str(err))
+                    elif ((url.endswith("mp4") or url.endswith("mov") or url.endswith("avi"))
+                          and self.session.http.head(url, raise_for_status=False).status_code == 200):
                         if quality_map is None:  # only make the request when it is necessary
                             quality_map = self._get_quality_map(content_id)
                         # rename the HTTP sources to match the HLS sources
