@@ -74,7 +74,11 @@ class WWENetwork(Plugin):
         data = self.session.http.json(res)
 
         if "status" in data and data["status"] != 200:
-            log.debug("API request failed: {0}:{1} ({2})".format(data["status"], data.get("code"), "; ".join(data.get("messages", []))))
+            log.debug("API request failed: {0}:{1} ({2})".format(
+                data["status"],
+                data.get("code"),
+                "; ".join(data.get("messages", []))
+            ))
         return data
 
     def login(self, email, password):
@@ -151,7 +155,11 @@ class WWENetwork(Plugin):
             self.logger.debug("Found content ID: {0}", content_id)
             info = self._get_media_info(content_id)
             if info.get("hlsUrl"):
-                for s in HLSStream.parse_variant_playlist(self.session, info["hlsUrl"], start_offset=start_point).items():
+                for s in HLSStream.parse_variant_playlist(
+                    self.session,
+                    info["hlsUrl"],
+                    start_offset=start_point
+                ).items():
                     yield s
             else:
                 log.error("Could not find the HLS URL")
